require 'bdd'

class VariableSet
  attr_reader :from, :to, :bdd_from, :bdd_to, :n

  def initialize(n)
    BDD.set_var_num(2*n)
    @n = n
    @from = (0..n-1).collect { |x| BDD.get_var(2*x) }
    @to = (0..n-1).collect { |x| BDD.get_var(2*x+1) }
    @bdd_from = BDD.make_set(@from)
    @bdd_to = BDD.make_set(@to)
    @succ = BDD::Pair.new
    @pred = BDD::Pair.new
    0.upto(n-1) { |i|
      @succ << [ @to[i], @from[i] ]
      @pred << [ @from[i], @to[i] ]
    }
  end

  def current2next
    @succ
  end

  def next2current
    @pred
  end

  def assigned(vars)
    res = BDD.true
    (@n-1).downto(0) { |i|
      res &= (@from[i] <=> @to[i]) if !vars.include?(i)
    }
    res
  end
end

def reachable_states(init,trans,var)
  r1 = init
  r2 = BDD.false
  until r1 == r2
    r2 = r1
#   r1 |= (r1 & trans).exist(var.bdd_from).replace(var.current2next)
    r1 |= r1.apply_exist(trans, BDD::AND, var.bdd_from).replace(var.current2next)
  end
  r1
end

#############

class Knight
  attr_reader :var, :trans, :init

  def initialize(n)
    @n = n
    @var = VariableSet.new(n*n)
    @trans = BDD.false
    @init = BDD.true
    0.upto(n-1) { |i|
      0.upto(n-1) { |j|
        @init &= (i == 0 && j == 0) ? x1(i,j) : x1(i,j).not
        [i-1,i+1].each { |k|
          [j-2,j+2].each { |l|
            @trans |= m(i,j,k,l) if on_board?(k,l)
          }
        }
        [i-2,i+2].each { |k|
          [j-1,j+1].each { |l|
            @trans |= m(i,j,k,l) if on_board?(k,l)
          }
        }
      }
    }
  end

  def on_board?(i,j)
    i >= 0 && j >= 0 && i < @n && j < @n
  end

  def m(i,j,k,l)
    x1(i,j) & x1(k,l).not & x2(i,j).not & x2(k,l) & assigned(x(i,j),x(k,l))
  end

  def visit_all
    BDD.nxor(@var.from)
  end

  def x(i,j)
    i*@n+j
  end
  def x1(i,j)
    @var.from[x(i,j)]
  end
  def x2(i,j)
    @var.to[x(i,j)]
  end
  def assigned(*v)
    @var.assigned(v)
  end

  def states
    return @reachable if @reachable
    @reachable = reachable_states(@init, @trans, @var)
  end

  def deadlocks
    return @deadlocks if @deadlocks
    @deadlocks = states & @trans.exist(@var.bdd_to).not
  end
end

k = Knight.new(8)
puts "Visit all: " + (k.states == k.visit_all).to_s
puts "Deadlocks: " + (k.deadlocks.false? ? "no" : "yes")
k.deadlocks.each{ |x| puts x.to_keys.inspect }
